/***************************************************************************
****   rdc_ax.c - by Alfonso Mangoni - Version 1.2 - 23.07.2002        *****
****************************************************************************

*****************************************************************************
rdc_ax.c uses input RDCs to find out the principal axis frame of the
molecule. i.e. the one where eq. 3 applies.
It performs a grid search over all the possible orientation of the molecule
and select the one where experimental and calculated RDCs fit best.
After this, the program can calculate any dipolar coupling in the molecule
as required by the command file.

rdc_ax.c needs only one (optional) argument, the name of the command file

the command file (default name: default.cmd) should have the following 
sintax (every line beginning with ; is a comment and is ignored):

The first line should be:

   source filename

where filename is the name of a PDB file with a 3D model of the  molecule.
Of course, the results are strongly dependent on the quality of this
structure.

Then, a section containing input RDCs begins with a line containing only the
keyword:

  experimental

followed by as many lines as the experimental RDCs used as input, with three
values separated by blank characters and indicating (1) the number of the first 
coupled atom as in the PBD file, (2) the the number of the second coupled atom,
and (3) the measured RDC, as in the following example:

  ;   ATOM1    ATOM2    COUPLING
      2         65         7.96      

This section ends with the keyword:

  end

Finally, a section containing RDCs to be calculated begins with a line 
containing only the keyword:

  calculated

followed by as many lines as the RDCs to be calculated, with two values 
separated by blank characters and indicating (1) the number of the first 
coupled atom as in the PBD file, (2) the the number of the second coupled 
atom, as in the following example:

  ;   ATOM1    ATOM2
      3         68         

This section ends with the keyword:

  end

Every line after this is ignored

***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#define PI_                     3.1415926535897932384626433832795
#define MAXCOUPLINGS            128


/****              global variables         ****/

int      nKnownCouplings;     /* Number of known dipolar coupling */

int      nUnknownCouplings;   /* Number of dipolar coupling to de calculated*/

double   dModR[MAXCOUPLINGS], /* Interatomic distances and components of    */ 
         dRx[MAXCOUPLINGS],   /* atoms whose dipolar coupling is known      */ 
         dRy[MAXCOUPLINGS], 
         dRz[MAXCOUPLINGS];

double   dModRc[MAXCOUPLINGS], /* Interatomic distances and components of    */ 
         dRcx[MAXCOUPLINGS],   /*atoms whose dipolar coupling has to be calcd*/ 
         dRcy[MAXCOUPLINGS], 
         dRcz[MAXCOUPLINGS];

float    fExpDipCoupl[MAXCOUPLINGS];     /* Experimental dipolar couplings    */
                                         /* (must be float because of sscanf) */

double   dCalcdDipCoupl[MAXCOUPLINGS];   /* Calculated couplings (after mult. */
                                         /* by dDax) for the orientation of   */
                                         /* the molecule being tested         */ 

double   dError;                         /* estimated RMS error of dipolar    */
                                         /* couplings for the orientation of  */
                                         /* the  molecule being tested        */

double   dDax;                           /* Dax (Equation 3) in orientation   */
                                         /* of the molecule being tested....  */
double   dDaxError;                       /* ... and its estimated RMS error  */    

double   dMinError;                      /* estimated RMS error of dipolar    */
                                         /* coupling   in the best orientation*/
                                         /* of the molecule                   */

int      nBestThetaX, nBestThetaY;       /* Rotation angles leading to the    */
                                         /* best orientation (integer)        */

double   dBestThetaX, dBestThetaY;       /* Rotation angles leading to the    */
                                         /* best orientation (double)         */

double   dBestDax;                       /* Dax (Equation 3) in the best      */
                                         /* orientation of the molecule....   */
double   dBestDaxError;                  /* ... and its RMS Error             */


/****              I/O variables             ****/

char   buffer[1024],                      /* Text buffers  */
       pdbfilename[1024];

/****             the main (and only) function             ****/


int main(int argc, char *argv[])
 {
   int      nThetaX,       /* Rotation angles for the current orientation */
            nThetaY;       

   int      nDipc;         /* Current dipolar coupling */

   double   dThetaX,       /* Rotation angles for the current orientation */
            dThetaY;
   
   double   dSinThetaX,    /* Sine and cosine of rotation angles */
            dCosThetaX, 
            dSinThetaY, 
            dCosThetaY;
   
   double   dRzPrime;      /* Component of the current bond along z */
   
   double   dCalDipCoupl;  /* The expression in parentheses in Eq. 3 for */
                           /* the current dipolar coupling                */
   
   /* These variables are for the least square calculations */
   double   dSumSqExpDipCoupl,    /* Sum of square experimental RDCs        */
            dSumSqCalcdDipCoupl,  /* Sum of square dCalDipCoupl's           */ 
            dSumExpCalcdDipCoupl; /* Sum of experim. RDCs * dCalDipCoupl's  */

   int      i;                 /* Loop counter */
   
   FILE     *pdbfile,          /* Pointer to the PDB file */
            *commandfile;      /* Pointer to the command file */
   
   int nLine;                  /* Current line number of the commanf file */
   
   char useless;               /* A meaningless argument for sscanf  */
   
   int nAtom1[MAXCOUPLINGS],   /* Atoms whose RDC is known */
       nAtom2[MAXCOUPLINGS];
   
   int nAtom1c[MAXCOUPLINGS],  /*Atoms whose RDC has to be calcd */
       nAtom2c[MAXCOUPLINGS];
   
   int bFoundEnd;               /* Boolean: has the keyword 'end' been found? */

   float fX1[MAXCOUPLINGS],   /* Cartesian coordinates of the two atoms     */
         fY1[MAXCOUPLINGS],   /* whose RDC is known                         */
         fZ1[MAXCOUPLINGS],
         fX2[MAXCOUPLINGS],
         fY2[MAXCOUPLINGS],
         fZ2[MAXCOUPLINGS];

   float fX1c[MAXCOUPLINGS],   /* Cartesian coordinates of the two atoms    */
         fY1c[MAXCOUPLINGS],   /* whose RDC has to be calcd                 */
         fZ1c[MAXCOUPLINGS],
         fX2c[MAXCOUPLINGS],
         fY2c[MAXCOUPLINGS],
         fZ2c[MAXCOUPLINGS];

   
/***** We want one and only one argument (the name of the command file) *****/

   if(argc>2)
    {
      puts("Usage: resdip commandfile");
      return 0;
    }

/***** We find no argument and use the default filename *****/

   if(argc==1)
    {
      commandfile=fopen("default.cmd", "r");
      if(commandfile==NULL)
      {
         printf("Cannot open source file; please provide a filename\n",
                                                             commandfile);
         return 0;
      }
    }

/***** We open the source file; if it fails, exit                       *****/

   commandfile=fopen(argv[1], "r");
   if(commandfile==NULL)
    {
      printf("Cannot open %s\n", argv[1]);
      return 0;
    }

/***** Scan the source file to find the name of the pdb file        *****/
/***** The first line of the source file must contain the source    *****/
/***** keyword                                                      *****/
   
   nLine=0;

   while(fgets(buffer, 1024, commandfile))
    {
      nLine++;
      if(buffer[0]==';')      /* lines beginning with ; are comments  */
         continue;

      if(sscanf(buffer, " source %s", pdbfilename)==0)
       {
         printf("%s: sintax error in line %d ('source' expected)\n",
                                                        argv[1], nLine);
         return 0;
       }
      break;
    }
   
   if(pdbfilename[0]==0)
    {
      printf("%s: no source section\n", argv[1]);
      return 0;
    }

/***** Now open the pdb file; if it fails, exit                   *****/

   pdbfile=fopen(pdbfilename, "r");
   if(pdbfile==NULL)
    {
      printf("Cannot open %s\n", pdbfilename);
      return 0;
    }

/***** SCan the source file: now the experimental couplings        *****/
/***** are expected                                                *****/
   
   while(fgets(buffer,1024,commandfile))
    {
      nLine++;
      if(buffer[0]==';')
         continue;

      if(sscanf(buffer, " experimental%c", &useless)==0)
       {
         printf("%s: sintax error in line %d ('experimental' expected)\n", 
                                                             argv[1], nLine);
         return 0;
       }
      break;
    }
   
/***** Scan the source file: each line should contain the         *****/
/***** number of two atoms and their measured coupling            *****/
/***** If the line contains 'end' we are done                     *****/
/***** These data are stored in the arrays nAtom1, nAtom2 (local) *****/
/***** and fExpDipCoupl   (global)                                *****/

   nKnownCouplings=0;
   bFoundEnd=0;
   while(fgets(buffer,1024,commandfile))
    {
      nLine++;
      if(buffer[0]==';')
         continue;

      if(sscanf(buffer, " end%c", &useless)!=0)
       {
         bFoundEnd=1;
         break;    
       }

      if(sscanf(buffer, " %d %d %f", &(nAtom1[nKnownCouplings]),
             &(nAtom2[nKnownCouplings]), &(fExpDipCoupl[nKnownCouplings]))==3)
       {
         nKnownCouplings++;
       }
      else
       {
         printf("%s: sintax error in line %d", argv[1], nLine);
         return 0;
       }
    }

   if(bFoundEnd==0)
    {
      printf("%s: unexpected end of file\n", argv[1], nLine);
      return 0;
    }
   
   
/***** Scan the source file: now the couplings to be                 *****/
/***** calculated are expected                                       *****/
   
   while(fgets(buffer,1024,commandfile))
    {
      nLine++;
      if(buffer[0]==';')
         continue;

      if(sscanf(buffer, " calculated%c", &useless)==0)
       {
         printf("%s: sintax error in line %d ('calculated' expected)\n", 
                                                             argv[1], nLine);
         return 0;
       }
      bFoundEnd=0;
      break;
    }
   

/***** Scan the source file: each line should contain the           *****/
/***** number of the two atoms whose coupling has to be calculated  *****/
/***** If the line contains 'end' we are done                       *****/
/***** These data are stored in the local arrays nAtom1c, nAtom2c   *****/

   nUnknownCouplings=0;
   while(fgets(buffer,1024,commandfile))
    {
      nLine++;
      if(buffer[0]==';')
         continue;

      if(sscanf(buffer, " end%c", &useless)!=0)
       {
         bFoundEnd=1;
         break;    
       }

      if(sscanf(buffer, " %d %d %f", &(nAtom1c[nUnknownCouplings]),
               &(nAtom2c[nUnknownCouplings]))==2)
       {
         nUnknownCouplings++;
       }
      else
       {
         printf("%s: sintax error in line %d", argv[1], nLine);
         return 0;
       }
    }

   if(bFoundEnd==0)
    {
      printf("%s: unexpected end of file\n", argv[1], nLine);
      return 0;
    }

/***** We are done with the source file and can close it            *****/

   fclose(commandfile);
   
/***** Now we know what to look for in the PDB file. Let's do it   *****/

   nLine=0;
   while(fgets(buffer,1024,pdbfile))
    {
      int nCurAtom;         /* The number of the current atom */
      float fX, fY, fZ;      /* The coordinates of the current atom */
      char buffer2[1024];

      if((sscanf(buffer, "ATOM%d%21c%f %f %f", 
                              &nCurAtom, &buffer2, &fX, &fY, &fZ)!=5) &&
               (sscanf(buffer, "HETATM%d%21c%f %f %f", 
                              &nCurAtom, &buffer2, &fX, &fY, &fZ)!=5))
         continue;

      for(i=0; i<nKnownCouplings; i++)
       {

         if(nAtom1[i]==nCurAtom && nAtom1[i]!=0)
          {
            fX1[i]=fX;
            fY1[i]=fY;
            fZ1[i]=fZ;
            nAtom1[i]=-nAtom1[i]; /* Change the sign if the atom is found */
          }

         if(nAtom2[i]==nCurAtom && nAtom2[i]!=0)
          {
            fX2[i]=fX;
            fY2[i]=fY;
            fZ2[i]=fZ;
            nAtom2[i]=-nAtom2[i]; /* Change the sign if the atom is found */
          }

       }

      for(i=0; i<nUnknownCouplings; i++)
       {

         if(nAtom1c[i]==nCurAtom && nAtom1c[i]!=0)
          {
            fX1c[i]=fX;
            fY1c[i]=fY;
            fZ1c[i]=fZ;
            nAtom1c[i]=-nAtom1c[i]; /* Change the sign if the atom is found */
          }
         
         if(nAtom2c[i]==nCurAtom && nAtom2c[i]!=0)
          {
            fX2c[i]=fX;
            fY2c[i]=fY;
            fZ2c[i]=fZ;
            nAtom2c[i]=-nAtom2c[i]; /* Change the sign if the atom is found */
          }
       
       }
   }
   
/***** We are done with the PDB file and can close it            *****/

   fclose(pdbfile);

   
/*****  Check if all atoms have been found                       ****/

   for(i=0; i<nKnownCouplings; i++)
    {
      if(nAtom1[i]>0)      /* This atom has not been found  */
       {
         printf("Atom %d not found in %s\n", nAtom1[i], pdbfilename);
         return(0);
       }
      if(nAtom2[i]>0)      /* This atom has not been found  */
       {
         printf("Atom %d not found in %s\n", nAtom2[i], pdbfilename);
         return(0);
       }
    }

   for(i=0; i<nUnknownCouplings; i++)
    {
      if(nAtom1c[i]>0)      /* This atom has not been found  */
       {
         printf("Atom %d not found in %s\n", nAtom1c[i], pdbfilename);
         return(0);
       }
      if(nAtom2c[i]>0)      /* This atom has not been found  */
       {
         printf("Atom %d not found in %s\n", nAtom2c[i], pdbfilename);
         return(0);
       }
    }

/**** Now can calculate the interatomic distances and their components ***/

   for(i=0; i<nKnownCouplings; i++)
    {
      dRx[i]=fX2[i]-fX1[i];
      dRy[i]=fY2[i]-fY1[i];
      dRz[i]=fZ2[i]-fZ1[i];
      dModR[i]=sqrt(dRx[i]*dRx[i]+dRy[i]*dRy[i]+dRz[i]*dRz[i]);
    }

   for(i=0; i<nUnknownCouplings; i++)
    {
      dRcx[i]=fX2c[i]-fX1c[i];
      dRcy[i]=fY2c[i]-fY1c[i];
      dRcz[i]=fZ2c[i]-fZ1c[i];
      dModRc[i]=sqrt(dRcx[i]*dRcx[i]+dRcy[i]*dRcy[i]+dRcz[i]*dRcz[i]);
    }


    
/**** We have to find the best orientation of the molecule. We do so by   *****              
***** trying every rotation angle (0-180 deg) along the x axis (nThetaX)  *****
***** and along the y axis (nThetaY).                                     ****/
   
   dMinError=1000.;          /* Just very large values */
   dBestDax=1000.;
   dBestDaxError=1000.;

   for(nThetaX=0; nThetaX<180; nThetaX++)
    {
      dThetaX=nThetaX*PI_/180;
      dCosThetaX=cos(dThetaX); /* These sine and cosine will be needed many */
      dSinThetaX=sin(dThetaX); /* times, so we calculate them only once here */
      for(nThetaY=0; nThetaY<180; nThetaY++)
       {
         dThetaY=nThetaY*PI_/180;
         dCosThetaY=cos(dThetaY);  /* See above */
         dSinThetaY=sin(dThetaY);

/**** For each couple of rotation angles the component along the z axis *****
***** (dRzPrime) of the vector between each couple of atoms is calcd.   *****
***** The experimental couplings and the ones expected on the basis of  *****
***** dRzPrime are correlated by the least squares method, according to *****
***** the relationship:                                                 *****
*****      fExpDipCoupl[n]=dDax*dCalDipCoupl                            *****
***** where fExpDipCoupl[n] is the n-th exp. coupling (treated as       *****
***** affected by experimental error), dCalDipCoupl is the expression   *****
***** in parentheses in eq. 3 (treated as exact) and dDax is the        *****
***** constant to be determined                                         ****/

         dSumSqExpDipCoupl=0;
         dSumSqCalcdDipCoupl=0;
         dSumExpCalcdDipCoupl=0;
         for(nDipc=0;nDipc<nKnownCouplings;nDipc++)
          { 
            dRzPrime=-dRx[nDipc]*dSinThetaY+
                         dRy[nDipc]*dSinThetaX*dCosThetaY+
                                   dRz[nDipc]*dCosThetaX*dCosThetaY;
            dCalDipCoupl=
                    (3*(dRzPrime/dModR[nDipc])*(dRzPrime/dModR[nDipc])-1);
            dSumSqExpDipCoupl=dSumSqExpDipCoupl+
                                     fExpDipCoupl[nDipc]*fExpDipCoupl[nDipc];
            dSumExpCalcdDipCoupl=dSumExpCalcdDipCoupl+
                                     fExpDipCoupl[nDipc]*dCalDipCoupl;
            dSumSqCalcdDipCoupl=dSumSqCalcdDipCoupl+
                                     dCalDipCoupl*dCalDipCoupl;
          }
         dDax=dSumExpCalcdDipCoupl/dSumSqCalcdDipCoupl;

/**** The least squares methods also provides the RMS error of   the        ****
***** exp. couplings (dError) and of dDax (dDaxError)                       ****
***** The latter is used to find out the orientation that best fits the     ****
***** experimental couplings                                                ***/
         
         dError=sqrt((dSumSqExpDipCoupl-dSumExpCalcdDipCoupl*dDax)/
                                                             nKnownCouplings);

         dDaxError=dError/sqrt(dSumSqCalcdDipCoupl);

         if(dDaxError <= dBestDaxError)  /* a better orientation! */
          {        
            dMinError=dError;
            nBestThetaX=nThetaX;
            nBestThetaY=nThetaY;
            dBestDax=dDax;
            dBestDaxError=dDaxError;
          }
       }
    }

/**** Now the unknown couplings are calculated....                        ****/
   
   dBestThetaX=nBestThetaX*PI_/180;
   dBestThetaY=nBestThetaY*PI_/180;
   for(nDipc=0;nDipc<nUnknownCouplings;nDipc++)
    { 
      dRzPrime=-dRcx[nDipc]*sin(dBestThetaY)+
                    dRcy[nDipc]*sin(dBestThetaX)*cos(dBestThetaY)+
                                dRcz[nDipc]*cos(dBestThetaX)*cos(dBestThetaY);
      dCalcdDipCoupl[nDipc]=dBestDax*
                    (3*(dRzPrime/dModRc[nDipc])*(dRzPrime/dModRc[nDipc])-1);
    }

/**** ....and the results are printed                                    ****/

   printf("Experimental (input) RDCs:\n");
   for(nDipc=0;nDipc<nKnownCouplings;nDipc++)
    { 
      printf("Atom %d - Atom %d:\t%+.2f Hz\n", 
                        -nAtom1[nDipc], -nAtom2[nDipc], fExpDipCoupl[nDipc]);
    }
   printf("\nD(ax): %+.2f%.2f Hz\n", 
                                    dBestDax, dBestDaxError);
   printf("ThetaX: %d degrees\n",nBestThetaX);
   printf("ThetaY %d degrees\n",nBestThetaY);
   printf("\nRMS Error: %+.2f Hz\n",dMinError);
   printf("\nCalculated RDCs:\n");
   for(nDipc=0;nDipc<nUnknownCouplings;nDipc++)
    { 
      printf("Atom %d - Atom %d:\t%+.2f Hz\n", 
                    -nAtom1c[nDipc], -nAtom2c[nDipc], dCalcdDipCoupl[nDipc]);
    }

   return 0;
 }
